#!/usr/bin/python2
import pygame
import sys
import json
import os
import re
try:
	import pygame._view # For py2exe
except:
	pass

pygame.init()

font = pygame.font.Font("Vera.ttf", 30)
smallfont = pygame.font.Font("Vera.ttf", 16)

images_path = 'img'
images = {}

TILE_W = 64
TILE_H = 64

#SCREEN_W = 800
#SCREEN_H = 600
SCREEN_W = 900
SCREEN_H = 680

def draw_shortinfo(screen):
	s = smallfont.render("Oh Snake by celeron55 @ GGJ2012", True, (100,100,100))
	screen.blit(s, (SCREEN_W-200, SCREEN_H-16))

def draw_gridbg(screen):
	for x in range(-1, int(SCREEN_W/TILE_W+2)):
		for y in range(-1, int(SCREEN_H/TILE_H+2)):
			screen.blit(get_image("bg"), (0+TILE_W*x, 0+TILE_H*y))

class Deserializer:
	def __init__(self):
		self.classes = {}
	def register(self, class_):
		self.classes[class_.sername] = class_
	def create(self, data):
		name = data['name']
		if name not in self.classes:
			raise GameError('Cannot deserialize '+repr(name))
		return self.classes[name].deserialize(data)
deserializer = Deserializer()

class Level:
	def __init__(self, filepath, name, worlddata):
		self.filepath = filepath
		self.name = name
		self.worlddata = worlddata
	def __str__(self):
		return ('Level('+repr(self.filepath)+', '+repr(self.name)+')')
	def save(self):
		#print("Writing level to "+repr(self.filepath))
		f = open(self.filepath, 'w+')
		data = {'name': self.name, 'worlddata': self.worlddata}
		json.dump(data, f, indent=4, sort_keys=True)
		f.close()

class LevelMgr:
	def __init__(self, levelpaths):
		self.levels = []
		self.levelpaths = levelpaths
		self.highestnumber = -1
	def load_all(self):
		for path in self.levelpaths:
			for filename in os.listdir(path):
				filepath = os.path.join(path, filename)
				if not re.match(r'\w+\.txt', filename):
					continue
				m = re.match(r'(\d+)\.txt', filename)
				if m:
					number = int(m.group(1))
					if number > self.highestnumber:
						self.highestnumber = number
				#print("Loading level from "+repr(filepath))
				f = open(filepath, 'r')
				data = json.load(f)
				f.close()
				level = Level(filepath, data['name'], data['worlddata'])
				#print("Loaded "+str(level))
				self.levels.append(level)
		def level_keygetter(level):
			m = re.match(r'.*(\d+)\.txt', level.filepath)
			try:
				if m:
					return int(m.group(1))
			except IndexError:
				pass
			return 999
		self.levels = sorted(self.levels, key=level_keygetter)
	def make_new(self, name):
		self.highestnumber += 1
		number = self.highestnumber
		filepath = os.path.join(self.levelpaths[0], str(number).zfill(2)+'.txt')
		level = Level(filepath, name, World(8,6).serialize())
		#print("Created "+str(level))
		self.levels.append(level)
		return level
		
levelmgr = LevelMgr(['levels'])
levelmgr.load_all()

class HighScoreMgr:
	def __init__(self, scorefilepath):
		self.scorefilepath = scorefilepath
		self.highscores = {}
	def load(self):
		try:
			f = open(self.scorefilepath, 'r')
			self.highscores = json.load(f)
			f.close()
		except IOError:
			pass
	def save(self):
		f = open(self.scorefilepath, 'w+')
		json.dump(self.highscores, f, indent=4, sort_keys=True)
		f.close()
	def set_score(self, levelfilepath, score):
		if score is None:
			try:
				del self.highscores[levelfilepath]
			except KeyError:
				pass
		else:
			self.highscores[levelfilepath] = score
		self.save()
	def get_score(self, levelfilepath):
		if levelfilepath in self.highscores:
			return self.highscores[levelfilepath]
		return None
highscoremgr = HighScoreMgr('highscores.txt')
highscoremgr.load()

class Section:
	def __init__(self):
		self.to_be_removed = False
	def on_draw(self):
		pass
	def on_pygame_event(self, event):
		pass
	def on_tick(self):
		pass

class SectionMgr:
	def __init__(self):
		self.stack = []
	def add(self, section):
		self.stack.append(section)
	def remove(self, section):
		self.stack.remove(section)
	def on_draw(self):
		for section in self.stack:
			if section.on_draw():
				break
	def on_pygame_event(self, event):
		for section in reversed(self.stack):
			if section.on_pygame_event(event):
				break
	def on_tick(self):
		for section in self.stack:
			if section.to_be_removed:
				self.stack.remove(section)
				continue
			section.on_tick()

sectionmgr = SectionMgr()

class GameError(Exception):
	def __init__(self, desc):
		self.desc = desc
	def __str__(self):
		return self.desc
	def __repr__(self):
		return self.desc

def get_image(filename):
	global images
	global images_path
	if filename in images:
		return images[filename]
	images[filename] = pygame.image.load(images_path + '/' + filename + '.png').convert_alpha()
	return images[filename]

def reverse_dir(d):
	return {'n':'s', 's':'n', 'w':'e', 'e':'w'}[d]

def dir_valid(d):
	return d in ['n','s','w','e']

def dir_vect(d):
	m = {'n':(0,-1), 's':(0,1), 'w':(-1,0), 'e':(1,0)}
	return m[d]

def vect_dir(x, y):
	if x == 0 and y == -1:
		return 'n'
	if x == 0 and y == 1:
		return 's'
	if x == -1 and y == 0:
		return 'w'
	if x == 1 and y == 0:
		return 'e'
	return None

def fix_dualdir(d1, d2):
	assert dir_valid(d1)
	assert dir_valid(d2)
	if d1 == d2:
		d2 = reverse_dir(d1)
	if d1 in ['w','e']:
		if d2 in ['w', 'e']:
			d1 = 'w'
			d2 = 'e'
		else:
			d1, d2 = d2, d1
	if d1 == 's' and d2 == 'n':
		d1, d2 = d2, d1
	return d1, d2

class Entity(object):
	def __init__(self, x, y):
		self.x = x
		self.y = y
		self.to_be_removed = False
	def get_image(self):
		return get_image('testimage')
	def get_text(self):
		return None
	def is_passable(self, from_x, from_y):
		return True
	def on_move_on(self, game, head):
		pass
	def pull(self, game, tox, toy, master_tox, master_toy):
		pass
	def tick(self, game):
		pass
	def edit_val_inc(self):
		pass
	def edit_val_dec(self):
		pass
	
class SnakeHead(Entity):
	def __init__(self, x, y, d_fwd):
		Entity.__init__(self, x, y)
		self.d_fwd = d_fwd
		assert dir_valid(d_fwd)
		self.nextpart = None
	def get_image(self):
		return get_image('snakehead_'+reverse_dir(self.d_fwd))
	def command_move(self, game, d):
		world = game.world
		dx, dy = dir_vect(d)
		newx, newy = self.x + dx, self.y + dy
		if not world.is_passable(newx, newy, self.x, self.y):
			# Rotate head freely if there is no tail
			if game.tail is self:
				self.d_fwd = d
			return
		game.tail_last_pos = (game.tail.x, game.tail.y)
		if self.nextpart:
			self.nextpart.pull(game, self.x, self.y, newx, newy)
		self.x = newx
		self.y = newy
		self.d_fwd = d
		for e in game.world.get_entities_at(self.x, self.y):
			e.on_move_on(game, self)
		game.score -= 1
	
	def is_move_possible(self, game, d):
		dx, dy = dir_vect(d)
		newx, newy = self.x + dx, self.y + dy
		return game.world.is_passable(newx, newy, self.x, self.y)

	def is_some_move_possible(self, game):
		for d in ['n','s','w','e']:
			if self.is_move_possible(game, d):
				return True
		return False
	
	sername = 'SnakeHead'
	@staticmethod
	def deserialize(data):
		return SnakeHead(int(data['x']), int(data['y']), data['d'])
	def serialize(self):
		return {'name':self.sername, 'x':self.x, 'y':self.y, 'd':self.d_fwd}
deserializer.register(SnakeHead)

class Border(Entity):
	def __init__(self, x, y, d):
		Entity.__init__(self, x, y)
		assert dir_valid(d)
		self.d = d
	def get_image(self):
		return get_image('border_'+self.d)
	def is_passable(self, from_x, from_y):
		return False

class Block(Entity):
	def __init__(self, x, y):
		Entity.__init__(self, x, y)
	def get_image(self):
		return get_image('block')
	def is_passable(self, from_x, from_y):
		return False

	sername = 'Block'
	@staticmethod
	def deserialize(data):
		return Block(int(data['x']), int(data['y']))
	def serialize(self):
		return {'name':self.sername, 'x':self.x, 'y':self.y}
deserializer.register(Block)

class FallingBlock(Entity):
	def __init__(self, x, y):
		Entity.__init__(self, x, y)
	def get_image(self):
		return get_image('falling_block')
	def is_passable(self, from_x, from_y):
		return False
	def tick(self, game):
		if not game.world.get_entities_at(self.x, self.y+1):
			self.y += 1

	sername = 'FallingBlock'
	@staticmethod
	def deserialize(data):
		return FallingBlock(int(data['x']), int(data['y']))
	def serialize(self):
		return {'name':self.sername, 'x':self.x, 'y':self.y}
deserializer.register(FallingBlock)

class Food(Entity):
	def __init__(self, x, y, imagename, points):
		Entity.__init__(self, x, y)
		self.imagename = imagename
		self.points = points
	def get_image(self):
		return get_image(self.imagename)
	def get_text(self):
		return str(self.points)
	def on_move_on(self, game, head):
		assert game.tail_last_pos
		tx, ty = game.tail_last_pos
		game.tail_last_pos = None
		self.to_be_removed = True
		newtail = SnakePart(tx, ty, 'w', 'e', None)
		game.world.add_entity(newtail)
		game.tail.nextpart = newtail
		game.tail = newtail

		game.score += self.points

	def edit_val_inc(self):
		self.points += 1
	def edit_val_dec(self):
		self.points -= 1

	sername = 'Food'
	@staticmethod
	def deserialize(data):
		return Food(int(data['x']), int(data['y']),
				data['imagename'], int(data['points']))
	def serialize(self):
		return {'name':self.sername, 'x':self.x, 'y':self.y,
				'imagename':self.imagename, 'points':self.points}
deserializer.register(Food)

class SnakePart(Entity):
	def __init__(self, x, y, d1, d2, nextpart):
		Entity.__init__(self, x, y)
		self.set_form(d1, d2)
		self.nextpart = nextpart
	def set_form(self, d1, d2):
		d1, d2 = fix_dualdir(d1, d2)
		self.ds = [d1, d2]
	def get_image(self):
		return get_image('snake_'+self.ds[0]+self.ds[1])
	def is_passable(self, from_x, from_y):
		return False
	def pull(self, game, tox, toy, master_tox, master_toy):
		dx1 = self.x - tox
		dy1 = self.y - toy
		d1 = vect_dir(dx1, dy1)
		dx2 = master_tox - tox
		dy2 = master_toy - toy
		d2 = vect_dir(dx2, dy2)
		if self.nextpart:
			self.nextpart.pull(game, self.x, self.y, tox, toy)
		self.set_form(d1, d2)
		self.x = tox
		self.y = toy

	def serialize(self):
		return {'name':'ExplodedSnake', 'x':self.x, 'y':self.y,
				'd1':self.ds[0], 'd2':self.ds[1]}

class ExplodedSnake(Entity):
	def __init__(self, x, y, d1, d2):
		Entity.__init__(self, x, y)
		d1, d2 = fix_dualdir(d1, d2)
		self.ds = [d1, d2]
	def get_image(self):
		return get_image('snake_'+self.ds[0]+self.ds[1])
	def is_passable(self, from_x, from_y):
		dx = from_x - self.x
		dy = from_y - self.y
		d = vect_dir(dx, dy)
		if d not in self.ds:
			return False
		return True
	def on_move_on(self, game, head):
		assert game.tail_last_pos
		tx, ty = game.tail_last_pos
		game.tail_last_pos = None
		self.to_be_removed = True
		newtail = SnakePart(tx, ty, 'w', 'e', None)
		game.world.add_entity(newtail)
		game.tail.nextpart = newtail
		game.tail = newtail
		
		in_x = game.head.nextpart.x
		in_y = game.head.nextpart.y
		in_dx = in_x - self.x
		in_dy = in_y - self.y
		in_d = vect_dir(in_dx, in_dy)
		assert in_d in self.ds
		if self.ds[0] == in_d:
			out_d = self.ds[1]
		else:
			out_d = self.ds[0]

		game.autofollow_timer = 3
		game.autofollow_d = out_d

		game.score += 2

	sername = 'ExplodedSnake'
	@staticmethod
	def deserialize(data):
		return ExplodedSnake(int(data['x']), int(data['y']),
				data['d1'], data['d2'])
	def serialize(self):
		return {'name':self.sername, 'x':self.x, 'y':self.y,
				'd1':self.ds[0], 'd2':self.ds[1]}
deserializer.register(ExplodedSnake)

class World:
	def __init__(self, w, h):
		self.w = w
		self.h = h
		self.entities = []
		#self.content = [[]]*w*h
		self.border_update()
	def get_entities_at(self, x, y):
		#return self.content[y*w+x]
		L = []
		for e in self.entities:
			if e.x == x and e.y == y:
				L.append(e)
		return L
	def clear_at(self, x, y):
		for entity in self.get_entities_at(x, y):
			entity.to_be_removed = True
	def add_entity(self, e):
		self.entities.append(e)
	def get_entities(self):
		return self.entities
	def is_passable(self, x, y, from_x, from_y):
		for e in self.get_entities_at(x, y):
			if not e.is_passable(from_x, from_y):
				return False
		return True
	def tick(self):
		for e in self.entities:
			if e.to_be_removed:
				self.entities.remove(e)
	def border_update(self):
		for e in self.entities:
			if isinstance(e, Border):
				e.to_be_removed = True
		for x in range(0, self.w):
			self.add_entity(Border(x, -1, 's'))
			self.add_entity(Border(x, self.h, 'n'))
		for y in range(0, self.h):
			self.add_entity(Border(-1, y, 'e'))
			self.add_entity(Border(self.w, y, 'w'))

	sername = 'World'
	@staticmethod
	def deserialize(data):
		w = World(int(data['w']), int(data['h']))
		for entitydata in data['entities']:
			entity = deserializer.create(entitydata)
			w.add_entity(entity)
		return w
	def serialize(self):
		data = {'name':self.sername, 'w':self.w, 'h':self.h,
				'entities':[]}
		for entity in self.entities:
			if hasattr(entity, 'serialize'):
				data['entities'].append(entity.serialize())
		return data
deserializer.register(World)

class Game:
	def __init__(self, world):
		self.world = world
		for e in world.get_entities():
			if isinstance(e, SnakeHead):
				self.head = e
				break
		else:
			raise GameError("head not found")
		self.tail = self.head
		self.tail_last_pos = None
		self.score = 0
		self.autofollow_d = None
		self.autofollow_timer = -1
	
	def tick(self):
		for e in self.world.get_entities():
			e.tick(self)
		self.world.tick()
	
def draw_world(world, screen, x0, y0):
	#for x in range(-2, world.w+2):
	#	for y in range(-2, world.h+2):
	for x in range(-4, world.w+4):
		for y in range(-4, world.h+4):
			screen.blit(get_image("bg"), (x0+TILE_W*x, y0+TILE_H*y))
	for e in world.get_entities():
		#print("drawing entity image "+repr(e.get_image())+" at "+str((x0+TILE_W*e.x, y0+TILE_H*e.y)))
		screen.blit(e.get_image(), (x0+TILE_W*e.x, y0+TILE_H*e.y))
		text = e.get_text()
		if text:
			s = font.render(text, False, (100,100,100))
			screen.blit(s, (int(x0+TILE_W*e.x+(TILE_W-s.get_width())/2),int(y0+TILE_H*e.y+6+(TILE_H-s.get_height())/2)))

def draw_score(game, screen, x0, y0):
	s = font.render("Score: "+str(game.score), False, (100,100,100))
	screen.blit(s, (x0,y0))

class BaseSection(Section):
	def __init__(self, screen):
		Section.__init__(self)
		self.screen = screen
	def on_draw(self):
		pass
	def on_pygame_event(self, event):
		if event.type == pygame.QUIT:
			sys.exit()

class MenuSection(Section):
	def __init__(self, screen):
		Section.__init__(self)
		self.screen = screen
		def play(self):
			sectionmgr.add(LevelSelectSection(self.screen, False))
		def editlevels(self):
			sectionmgr.add(LevelSelectSection(self.screen, True))
		def credits(self):
			sectionmgr.add(CreditsSection(self.screen))
		def quit(self):
			sys.exit()
		self.options = [
			["Play", play],
			["Edit levels", editlevels],
			["Credits", credits],
			["Quit", quit]
		]
		self.cursor = 0
	def on_draw(self):
		#self.screen.fill((255,255,255))
		draw_gridbg(self.screen)
		self.screen.blit(get_image("splash"), (int((SCREEN_W-640)/2),
				int((SCREEN_H-480)/2)))
		i = 0
		for (name, func) in self.options:
			text = ""
			if i == self.cursor:
				text += '>'
			text += name
			s = font.render(text, False, (0,0,0))
			screen.blit(s, (SCREEN_W/2-50,SCREEN_H/2 + 30*i))
			i += 1
		draw_shortinfo(screen)
	def on_pygame_event(self, event):
		if event.type == pygame.KEYDOWN:
			if event.key == pygame.K_UP:
				if self.cursor > 0:
					self.cursor -= 1
				else:
					self.cursor = len(self.options)-1
			elif event.key == pygame.K_DOWN:
				if self.cursor < len(self.options)-1:
					self.cursor += 1
				else:
					self.cursor = 0
			elif event.key in [pygame.K_RETURN, pygame.K_SPACE]:
				self.options[self.cursor][1](self)
			elif event.key == pygame.K_ESCAPE:
				sys.exit()
			return True

class CreditsSection(Section):
	def __init__(self, screen):
		Section.__init__(self)
		self.screen = screen
	def on_draw(self):
		#self.screen.fill((255,255,255))
		draw_gridbg(self.screen)
		texts = [
			'Oh Snake',
			'',
			'Made by Perttu "celeron55" Ahola',
			'At GGJ/FGJ Oulu 2012'
		]
		i = 0
		for text in texts:
			s = font.render(text, False, (0,0,0))#, (255,255,255))
			screen.blit(s, (200,230 + 30*i))
			i += 1
	def on_pygame_event(self, event):
		if event.type == pygame.KEYDOWN:
			if event.key in [pygame.K_RETURN, pygame.K_SPACE, pygame.K_ESCAPE]:
				self.to_be_removed = True
			return True

class ShowMessageSection(Section):
	def __init__(self, screen, message):
		Section.__init__(self)
		self.screen = screen
		self.message = message
	def on_draw(self):
		#self.screen.fill((255,255,255))
		draw_gridbg(self.screen)
		texts = self.message.split('\n')
		i = 0
		for text in texts:
			s = font.render(text, False, (0,0,0))#, (255,255,255))
			screen.blit(s, (150,200 + 30*i))
			i += 1
	def on_pygame_event(self, event):
		if event.type == pygame.KEYDOWN:
			if event.key in [pygame.K_RETURN, pygame.K_SPACE, pygame.K_ESCAPE]:
				self.to_be_removed = True
			return True

class LevelSelectSection(Section):
	def __init__(self, screen, do_edit):
		Section.__init__(self)
		self.screen = screen
		self.do_edit = do_edit
		self.cursor = 0
		self.refresh_options()
	def refresh_options(self):
		self.options = []
		i = 0
		for level in levelmgr.levels:
			self.options.append({
				'text': str(i+1)+'. '+level.name,
				'score': highscoremgr.get_score(level.filepath),
				'level': level
			})
			i += 1
	def on_draw(self):
		draw_gridbg(self.screen)
		self.refresh_options()
		text = "Available levels:"
		if self.do_edit:
			text += "  (press N to make a new one)"
		s = font.render(text, False, (0,0,0))
		screen.blit(s, (100,100))
		i = 0
		for optline in self.options:
			name = optline['text']
			score = optline['score']
			text = ""
			if i == self.cursor:
				text += '>'
			text += name
			s = font.render(text, False, (0,0,0))
			screen.blit(s, (150, 150 + 30*i))
			scoretext = '-'
			if score is not None:
				scoretext = str(score)
			s = font.render(str(scoretext), False, (0,0,0))
			screen.blit(s, (650, 150 + 30*i))
			i += 1
		draw_shortinfo(screen)
	def on_pygame_event(self, event):
		if event.type == pygame.KEYDOWN:
			if event.key == pygame.K_UP:
				if self.cursor > 0:
					self.cursor -= 1
				else:
					self.cursor = len(self.options)-1
			elif event.key == pygame.K_DOWN:
				if self.cursor < len(self.options)-1:
					self.cursor += 1
				else:
					self.cursor = 0
			elif event.key in [pygame.K_RETURN, pygame.K_SPACE]:
				level = self.options[self.cursor]['level']
				if self.do_edit:
					gs = EditorSection(self.screen, level)
					sectionmgr.add(gs)
				else:
					gs = GameSection(self.screen, level)
					sectionmgr.add(gs)
			if event.key in [pygame.K_ESCAPE]:
				self.to_be_removed = True
			if self.do_edit:
				if event.key == pygame.K_n:
					level = levelmgr.make_new("No name")
					gs = EditorSection(self.screen, level)
					sectionmgr.add(gs)
			return True

class EditorSection(Section):
	def __init__(self, screen, level):
		Section.__init__(self)
		self.screen = screen
		self.level = level
		self.world = deserializer.create(level.worlddata)
		self.cursor_x = int(self.world.w/2)
		self.cursor_y = int(self.world.h/2)
	def on_draw(self):
		screen = self.screen
		world = self.world
		world_x = int((screen.get_width() - TILE_W*(world.w))/2)
		world_y = int(TILE_H+(screen.get_height()-TILE_H - TILE_H*world.h)/2)
		draw_world(world, screen, world_x, world_y)
		# Draw cursor
		screen.blit(get_image("cursor"), (world_x+TILE_W*self.cursor_x-2,
				world_y+TILE_H*self.cursor_y-2))
		texts = [
			'Snakes: QW,AS: corner, ED: straight, R: head',
			'Empty: Z, Food: F, Block: X, FallingBlock: C',
			'Values: PGUP/PGDOWN, Level size: Shift+Arr.',
			'Save: Ctrl+S, Modify level name: Ctrl+N'
		]
		i = 0
		for text in texts:
			s = font.render(text, False, (0,0,0))#, (255,255,255))
			screen.blit(s, (30, 10 + 30*i))
			i += 1
		texts = [
			'Level name: "'+self.level.name+'" ('+self.level.filepath+')'
		]
		i = 0
		for text in texts:
			s = font.render(text, False, (0,0,0))#, (255,255,255))
			screen.blit(s, (30, SCREEN_H-s.get_height()*len(texts) + 30*i))
			i += 1
		draw_shortinfo(screen)
		
	def on_pygame_event(self, event):
		world = self.world
		if event.type == pygame.KEYDOWN:
			if event.mod in [pygame.KMOD_LCTRL, pygame.KMOD_CTRL,
					pygame.KMOD_RCTRL]:
				if event.key == pygame.K_s:
					self.level.worlddata = world.serialize()
					self.level.save()
					highscoremgr.set_score(self.level.filepath, None)
					return
				if event.key == pygame.K_n:
					gs = SetLevelNameSection(self.screen, self.level)
					sectionmgr.add(gs)

			keydirmap = {
				pygame.K_LEFT: 'w',
				pygame.K_RIGHT: 'e',
				pygame.K_UP: 'n',
				pygame.K_DOWN: 's'
			}
			if event.key in keydirmap:
				d = keydirmap[event.key]
				dx, dy = dir_vect(d)
				if event.mod in [pygame.KMOD_LSHIFT, pygame.KMOD_SHIFT,
						pygame.KMOD_RSHIFT]:
					self.world.w += dx
					self.world.h += dy
				else:
					self.cursor_x += dx
					self.cursor_y += dy
			if event.key in [pygame.K_ESCAPE]:
				self.to_be_removed = True

			snakepiecemap = {
				pygame.K_q: ['s','e'],
				pygame.K_w: ['s','w'],
				pygame.K_a: ['n','e'],
				pygame.K_s: ['n','w'],
				pygame.K_e: ['n','s'],
				pygame.K_d: ['w','e'],
			}
			if event.key in snakepiecemap:
				world.clear_at(self.cursor_x, self.cursor_y)
				dirs = snakepiecemap[event.key]
				world.add_entity(ExplodedSnake(self.cursor_x, self.cursor_y,
						dirs[0], dirs[1]))

			staticpiecemap = {
				pygame.K_r: SnakeHead(self.cursor_x, self.cursor_y, 'w'),
				pygame.K_z: None,
				pygame.K_x: Block(self.cursor_x, self.cursor_y),
				pygame.K_c: FallingBlock(self.cursor_x, self.cursor_y),
				pygame.K_f: Food(self.cursor_x, self.cursor_y, "apple", 10),
			}
			if event.key in staticpiecemap:
				world.clear_at(self.cursor_x, self.cursor_y)
				piece = staticpiecemap[event.key]
				if piece:
					world.add_entity(piece)

			if event.key == pygame.K_PAGEUP:
				for e in world.get_entities_at(self.cursor_x, self.cursor_y):
					e.edit_val_inc()
			if event.key == pygame.K_PAGEDOWN:
				for e in world.get_entities_at(self.cursor_x, self.cursor_y):
					e.edit_val_dec()

			# Update borders everytime because they can be edited and
			# map size may change
			self.world.border_update()

			return True
	def on_tick(self):
		self.world.tick()

class SetLevelNameSection(Section):
	def __init__(self, screen, level):
		Section.__init__(self)
		self.screen = screen
		self.level = level
		self.name = level.name
		self.counter = 0
	def on_draw(self):
		screen = self.screen
		draw_gridbg(self.screen)
		blinker = ''
		if self.counter % 10 < 5:
			blinker = '_'
		texts = [
			'Input name:',
			self.name+blinker,
		]
		i = 0
		for text in texts:
			s = font.render(text, False, (0,0,0))#, (255,255,255))
			screen.blit(s, (150,200 + 30*i))
			i += 1
		draw_shortinfo(screen)
	def on_pygame_event(self, event):
		if event.type == pygame.KEYDOWN:
			if event.key == pygame.K_RETURN:
				self.level.name = self.name
				self.level.save()
				self.to_be_removed = True
			elif event.key == pygame.K_ESCAPE:
				self.to_be_removed = True
			elif event.key == pygame.K_BACKSPACE:
				self.name = self.name[0:-1]
			else:
				self.name += event.unicode
			return True
	def on_tick(self):
		self.counter += 1

class GameSection(Section):
	def __init__(self, screen, level):
		Section.__init__(self)
		self.screen = screen
		self.level = level
		world = deserializer.create(level.worlddata)
		try:
			self.game = Game(world)
		except GameError:
			s = ShowMessageSection(self.screen, "Error setting up game")
			sectionmgr.add(s)
			self.to_be_removed = True
		self.delayed_levelendaction = None
	def on_draw(self):
		screen = self.screen
		game = self.game
		world_x = int((screen.get_width() - TILE_W*game.world.w)/2)
		world_y = int((screen.get_height() - TILE_H*game.world.h)/2)
		draw_world(game.world, screen, world_x, world_y)
		draw_score(game, screen, world_x, world_y - 40)
		draw_shortinfo(screen)
	def on_pygame_event(self, event):
		game = self.game
		if event.type == pygame.KEYDOWN:
			if not game.autofollow_timer != -1:
				keydirmap = {
					pygame.K_LEFT: 'w',
					pygame.K_RIGHT: 'e',
					pygame.K_UP: 'n',
					pygame.K_DOWN: 's'
				}
				if event.key in keydirmap:
					d = keydirmap[event.key]
					game.head.command_move(game, d)
			if event.key in [pygame.K_ESCAPE]:
				self.to_be_removed = True
			return True
	def on_tick(self):
		game = self.game
		# End game if last tick caused the level to end
		if self.delayed_levelendaction is not None:
			gs = LevelEndSection(self.screen, self.game,
					self.delayed_levelendaction == "success",
					self.level)
			sectionmgr.add(gs)
			self.to_be_removed = True
			return
		# Check end conditions
		exploded_left = False
		for e in game.world.get_entities():
			if isinstance(e, ExplodedSnake):
				exploded_left = True
				break
		levelendaction = None
		if not exploded_left:
			levelendaction = "success"
		if not game.head.is_some_move_possible(game):
			levelendaction = "fail"
		if levelendaction is not None:
			self.delayed_levelendaction = levelendaction
		# Run stuff
		if game.autofollow_timer != -1:
			game.autofollow_timer -= 1
			if game.autofollow_timer <= 0:
				game.autofollow_timer = -1
				game.head.command_move(game, game.autofollow_d)
		game.tick()

class LevelEndSection(Section):
	def __init__(self, screen, game, success, level):
		Section.__init__(self)
		self.screen = screen
		self.game = game
		self.success = success
		s = screen.copy().convert_alpha()
		self.gameshot = pygame.transform.rotozoom(s, -10, 0.33)
		self.best_score = False
		if success:
			highscore = highscoremgr.get_score(level.filepath)
			if highscore is None or game.score > highscore:
				self.best_score = True
				print("Saving highscore for "+str(level))
				highscoremgr.set_score(level.filepath, game.score)
	def on_draw(self):
		global font
		screen = self.screen
		game = self.game
		#self.screen.fill((255,255,255))
		draw_gridbg(self.screen)
		if self.success:
			screen.blit(get_image("level_success"), (int((SCREEN_W-640)/2),
					int((SCREEN_H-480)/2)))
			screen.blit(self.gameshot, ((SCREEN_W)/2, (SCREEN_H)/2))
			bestscoretext = ""
			if self.best_score:
				bestscoretext = " (Best Score!)"
			s = font.render(" Score: "+str(game.score)+" "+bestscoretext,
					False, (0,0,0),
					(255,255,255))
			screen.blit(s, (200,300))
		else:
			screen.blit(get_image("level_fail"), (int((SCREEN_W-640)/2), int((SCREEN_H-480)/2)))
			screen.blit(self.gameshot, ((SCREEN_W)/2, (SCREEN_H)/2))
		draw_shortinfo(screen)
	def on_pygame_event(self, event):
		if event.type == pygame.KEYDOWN:
			self.to_be_removed = True
			return True

screen_w = SCREEN_W
screen_h = SCREEN_H
screen = pygame.display.set_mode((screen_w, screen_h))

pygame.display.set_caption('Oh Snake')
pygame.mouse.set_visible(0)

sectionmgr.add(BaseSection(screen))
sectionmgr.add(MenuSection(screen))

while 1:
	for event in pygame.event.get():
		sectionmgr.on_pygame_event(event)
	sectionmgr.on_tick()
	sectionmgr.on_draw()
	pygame.display.update()
	pygame.time.delay(50)

# END
